unit ShxFont;

interface

uses
  Winapi.Windows, System.SysUtils, System.Classes, System.Contnrs, System.Math,
  System.UITypes;

type
  //  
  TShxFontKind = (fkShpFont, fkUniFont, fkBigFont, fkUnknown);

  TShxFont = class;

  // 
  TShxShape = class(TObject)
  private
    FFont: TShxFont;
    FNumber: Word;
    FBytesCount: Word;
    FName: String;
    FGlyphHighIndex: Word;
  protected
    FGlyph: array of Byte;
    function GetGlyph(Index: Integer): Byte;
    procedure SetGlyph(Index: Integer; const Value: Byte);
    procedure SetGlyphLength(const Value: Word);
    function GetGlyphLength: Word;
  public
    constructor Create(AFont: TShxFont; ANumber, ABytesCount: Word);
    destructor Destroy; override;
    function GetGlyphDefinition(AMaxStrLen: Byte): String;
    property Number: Word read FNumber write FNumber;
    property BytesCount: Word read FBytesCount write FBytesCount;
    property Name: String read FName write FName;
    property Glyph[Index: Integer]: Byte read GetGlyph write SetGlyph;
    property GlyphLength: Word read GetGlyphLength write SetGlyphLength;
  end;

  // 
  TShxFont = class(TObjectList)
  private
    FFontFileName: String;   //     (-   LoadFromFile)
    FFontName: String;       // =   0 (    )
    FFontKind: TShxFontKind; //  
    FIsFont: Boolean;        //  " ": True  , False   
    FAbove: Byte;            //      
    FBelow: Byte;            //      
    FModes: Byte;            //   
    FUniEncoding: Byte;      //  Unicode
    FUniType: Byte;          //  
    FMaxStrLen: Byte;        // .     (   SHP)
  protected
    function GetShape(Index: Integer): TShxShape;
    procedure SetShape(Index: Integer; ASample: TShxShape);
    procedure SetMaxStrLen(const Value: Byte);
    procedure SetDefault;
    procedure LoadShape(const FileHandle: THandle; const ShapeNum: Word);
  public
    constructor Create;
    function Add(AShape: TShxShape): Integer;
    procedure Insert(Index: Integer; ASample: TShxShape);
    procedure ExportToShpFile(const AFileName: String);
    function LoadFromFile(const AFileName: String): Boolean;
    function GetShapeIndex(ShapeNumber: Word): Integer;
    property Shapes[Index: Integer]: TShxShape read GetShape write SetShape; default;
    property FontFileName: String read FFontFileName;
    property FontName: String read FFontName;
    property FontKind: TShxFontKind read FFontKind;
    property IsFont: Boolean read FIsFont;
    property Above: Byte read FAbove;
    property Below: Byte read FBelow;
    property Modes: Byte read FModes;
    property UniEncoding: Byte read FUniEncoding;
    property UniType: Byte read FUniType;
    property MaxStrLen: Byte read FMaxStrLen write SetMaxStrLen;
  end;

  function ShxFontKindToStr(AFontKind: TShxFontKind): String;

var
  FontShx: TShxFont;

implementation

type
  TShxShapeInfo = packed record
    ShapeNumber: Word;
    GlyphBytes: Word;
  end;

  THackedShxShape = class(TShxShape);

const
  MAX_STR_LEN_MIN: Byte = 8;
  MAX_STR_LEN_MAX: Byte = 128;

resourcestring
  rsShpFont = '';
  rsUniFont = '';
  rsBigFont = '';
  rsUnknown = '';

function GetFontKind(AFileHeader: AnsiString): TShxFontKind;
const
  SHX_SHAPES10: AnsiString = 'AutoCAD-86 shapes 1.0';
  SHX_SHAPES11: AnsiString = 'AutoCAD-86 shapes 1.1';
  SHX_UNIFONT10: AnsiString = 'AutoCAD-86 unifont 1.0';
  SHX_BIGFONT10: AnsiString = 'AutoCAD-86 bigfont 1.0';
begin
  if (Pos(SHX_SHAPES10, AFileHeader) <> 0) or (Pos(SHX_SHAPES11, AFileHeader) <> 0) then
    Result:= fkShpFont
  else if (Pos(SHX_UNIFONT10, AFileHeader) <> 0) then
    Result:= fkUniFont
  else if (Pos(SHX_BIGFONT10, AFileHeader) <> 0) then
    Result:= fkBigFont
  else
    Result:= fkUnknown;
end;

function ShxFontKindToStr(AFontKind: TShxFontKind): String;
begin
  case AFontKind of
    fkShpFont: Result:= rsShpFont;
    fkUniFont: Result:= rsUniFont;
    fkBigFont: Result:= rsBigFont;
  else
    Result:= rsUnknown;
  end;
end;

function EncodeOctant(DSC: Byte): String;
var
  S, C: Byte;
  CW: Boolean;
  tmp: Shortint;
  b: Byte;
begin
  tmp:= Shortint(DSC);
  CW:= ((tmp and $80) <> 0);
  b:= Byte(tmp);
  S:= (b and $70) shr 4;
  C:= b and $07;
  if CW then
    Result:= Format('-0%.1x%.1x', [S, C])
  else
    Result:= Format('0%.1x%.1x', [S, C]);
end;

{ TShxShape }

constructor TShxShape.Create(AFont: TShxFont; ANumber, ABytesCount: Word);
begin
  FFont:= AFont;
  FNumber:= ANumber;
  FBytesCount:= ABytesCount;
end;

destructor TShxShape.Destroy;
begin
  SetLength(FGlyph, 0);
  inherited Destroy;
end;

function TShxShape.GetGlyph(Index: Integer): Byte;
begin
  Result:= FGlyph[Index];
end;

function TShxShape.GetGlyphDefinition(AMaxStrLen: Byte): String;
var
  i: Integer;
  n, MaxLen: Byte;
begin
  MaxLen:= EnsureRange(AMaxStrLen, MAX_STR_LEN_MIN, MAX_STR_LEN_MAX);
  Result:= Format('*%d,%d', [FNumber, GlyphLength]);
  if FName = '' then
    Result:= Result + #13#10
  else
    Result:= Format('%s,%s'#13#10, [Result, FName]);
  n:= 0;
  FGlyphHighIndex:= High(FGlyph);
  i:= Low(FGlyph);
  while i <= FGlyphHighIndex do
  begin
    case FGlyph[i] of
      $0: Result:= Format('%s0', [Result]); //   
      $1: Result:= Format('%s%d,', [Result, FGlyph[i]]); //  
      $2: Result:= Format('%s%d,', [Result, FGlyph[i]]); //  
      $3: begin //      
            Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]]);
            Inc(i);
            Inc(n);
          end;
      $4: begin //      
            Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]]);
            Inc(i);
            Inc(n);
          end;
      //     
      $5: Result:= Format('%s%d,', [Result, FGlyph[i]]);
      //     
      $6: Result:= Format('%s%d,', [Result, FGlyph[i]]);
      $7: begin //  ,     
            if FFont.FFontKind = fkUniFont then // ( UNIFONT - )
            begin
              Result:= Format('%s%d,0%.4x,', [Result, FGlyph[i],
                  (FGlyph[i + 1] shl 8) + FGlyph[i + 2]]);
              Inc(i, 2);
              Inc(n, 2);
            end else
            begin
              Result:= Format('%s%d,%d,', [Result, FGlyph[i], FGlyph[i + 1]]);
              Inc(i);
              Inc(n);
            end;
          end;
      $8: begin //    X-Y,    
            Result:= Format('%s%d,(%d,%d),', [Result, FGlyph[i],
                Shortint(FGlyph[i + 1]), Shortint(FGlyph[i + 2])]);
            Inc(i, 2);
            Inc(n, 2);
          end;
      $9: begin //       X-Y,   (0,0)
            Result:= Format('%s%d,', [Result, FGlyph[i]]);
            Inc(n);
            Inc(i, 2); //    dY
            while True do
            begin
              Result:= Format('%s(%d,%d),', [Result,
                Shortint(FGlyph[i - 1]), Shortint(FGlyph[i])]);
              Inc(n, 2);
              if n > MaxLen then
              begin
                Result:= Result + #13#10;
                n:= 0;
              end;
              if (FGlyph[i - 1] = 0) and (FGlyph[i] = 0) then // [(dX,dY) = (0,0)]?
                Break;
              Inc(i, 2);
            end;
          end;
      $A: begin //  ,    
            Result:= Format('%s%d,(%d,%s),', [Result, FGlyph[i], FGlyph[i + 1],
                EncodeOctant(FGlyph[i + 2])]);
            Inc(i, 2);
            Inc(n, 2);
          end;
      $B: begin //  ,   5 
            Result:= Format('%s%d,(%d,%d,%d,%d,%s),', [Result,
                FGlyph[i], FGlyph[i + 1], FGlyph[i + 2], FGlyph[i + 3], FGlyph[i + 4],
                EncodeOctant(FGlyph[i + 5])]);
            Inc(i, 5);
            Inc(n, 5);
          end;
      $C: begin // ,   X-Y  
            Result:= Format('%s%d,(%d,%d,%d),', [Result, FGlyph[i],
              Shortint(FGlyph[i + 1]), Shortint(FGlyph[i + 2]), Shortint(FGlyph[i + 3])]);
            Inc(i, 3);
            Inc(n, 3);
          end;
      $D: begin //  ,   X-Y  
            Result:= Format('%s%d,', [Result, FGlyph[i]]);
            Inc(i, 3); //    Bulge
            while True do
            begin
              if (FGlyph[i - 2] = 0) and (FGlyph[i - 1] = 0) then // [(dX,dY) = (0,0)]?
              begin
                Result:= Format('%s(0,0),', [Result]);
                Dec(i);
                Break;
              end else
              begin
                Result:= Format('%s(%d,%d,%d),', [Result,
                  Shortint(FGlyph[i - 2]), Shortint(FGlyph[i - 1]), Shortint(FGlyph[i])]);
                if n > MaxLen then
                begin
                  Result:= Result + #13#10;
                  n:= 0;
                end;
              end;
              Inc(i, 3);
              Inc(n, 3);
            end;
          end;
      //       
      $E: Result:= Format('%s%d,', [Result, FGlyph[i]]);
    else
      //     
      Result:= Format('%s0%.2x,', [Result, FGlyph[i]]);
    end;
    Inc(n);
    if n > MaxLen then
    begin
      Result:= Result + #13#10;
      n:= 0;
    end;
    Inc(i);
  end;
end;

function TShxShape.GetGlyphLength: Word;
begin
  Result:= Length(FGlyph);
end;

procedure TShxShape.SetGlyph(Index: Integer; const Value: Byte);
begin
  FGlyph[Index]:= Value;
end;

procedure TShxShape.SetGlyphLength(const Value: Word);
begin
  SetLength(FGlyph, Value);
end;

{ TShxShapes }

function TShxFont.Add(AShape: TShxShape): Integer;
begin
  Result:= inherited Add(AShape);
end;

constructor TShxFont.Create;
begin
  inherited Create;
  SetDefault;
  FMaxStrLen:= 10;
  FFontKind:= fkUnknown;
end;

procedure TShxFont.ExportToShpFile(const AFileName: String);
var
  Lst: TStringList;
  i: Integer;
begin
  if Count < 1 then Exit;
  Lst:= TStringList.Create;
  try
    if FFontKind = fkShpFont then
    begin
      Lst.Add(Format('*0,4,%s', [FFontName]));
      Lst.Add(Format('%d,%d,%d,0', [FAbove, FBelow, FModes]));
    end else // fkUniFont
    begin
      Lst.Add(Format('*UNIFONT,6,%s', [FFontName]));
      Lst.Add(Format('%d,%d,%d,%d,%d,0', [FAbove, FBelow, FModes, FUniEncoding, FUniType]));
    end;
    for i:= 0 to Count - 1 do
      Lst.Add(Shapes[i].GetGlyphDefinition(FMaxStrLen));
    Lst.Add('');
    Lst.SaveToFile(AFileName);
  finally
    Lst.Free;
  end;
end;

function TShxFont.GetShape(Index: Integer): TShxShape;
begin
  Result:= TShxShape(inherited Items[Index]);
end;

function TShxFont.GetShapeIndex(ShapeNumber: Word): Integer;
var
  First, Last, Mid: Integer;
begin
  Result:= -1;
  if Count = 0 then Exit;
  First:= 0;
  Last:= Count - 1;
  //  
  while (First < Last) do
  begin
    Mid:= First + (Last - First) div 2;
    if Shapes[Mid].Number >= ShapeNumber then
      Last:= Mid
    else
      First:= Mid + 1;
  end;
  if Shapes[Last].Number = ShapeNumber then
    Result:= Last;
end;

procedure TShxFont.SetDefault;
begin
  FIsFont:= False;
  FAbove:= 10;
  FBelow:= 3;
  FModes:= 0;
  FUniEncoding:= 2;
  FUniType:= 0;
  FFontName:= '';
  FFontFileName:= '';
end;

procedure TShxFont.SetMaxStrLen(const Value: Byte);
begin
  FMaxStrLen:= EnsureRange(Value, MAX_STR_LEN_MIN, MAX_STR_LEN_MAX);
end;

procedure TShxFont.SetShape(Index: Integer; ASample: TShxShape);
begin
  inherited Items[Index]:= ASample;
end;

procedure TShxFont.Insert(Index: Integer; ASample: TShxShape);
begin
  inherited Insert(Index, ASample);
end;

procedure TShxFont.LoadShape(const FileHandle: THandle; const ShapeNum: Word);
var
  GlyphLen, GlyphOfs: Word;
  Buffer: PAnsiChar;
  S: AnsiString;
begin
  Buffer:= System.AllocMem(Shapes[ShapeNum].BytesCount + 1);
  try
    FileRead(FileHandle, Buffer^, Shapes[ShapeNum].BytesCount);
    S:= AnsiString(Buffer);
    Shapes[ShapeNum].Name:= String(S);
    //  
    GlyphOfs:= Length(Shapes[ShapeNum].Name) + 1;
    GlyphLen:= Shapes[ShapeNum].BytesCount - GlyphOfs;
    Shapes[ShapeNum].GlyphLength:= GlyphLen;
    System.Move(Buffer[GlyphOfs], THackedShxShape(Shapes[ShapeNum]).FGlyph[0], GlyphLen);
  finally
    FreeMem(Buffer);
  end;
end;

function TShxFont.LoadFromFile(const AFileName: String): Boolean;
const
  SHX_FILE_SIGNATURE_SIZE: Word = 32;     // ,     
  SHX_SHAPES_NUM_SHAPES_OFS: Word = $1C;  //   .    
  SHX_UNIFONT_NUM_SHAPES_OFS: Word = $19;
  SHX_SHAPES_SHAPE0_GLYPH_LENGTH: Word = $03; //       0
  SHX_UNIFNT_SHAPE0_GLYPH_LENGTH: Word = $05;
var
  FileHandle: THandle;
  NumShapesOfs: Integer;
  i, NumShapes, Shape0GlyphLen: Word;
  ShapeInfo: TShxShapeInfo;
  Buffer: PAnsiChar;
  S: AnsiString;
begin
  Result:= False;
  Clear;
  FileHandle:= FileOpen(AFileName, fmOpenRead);
  try
    FileSeek(FileHandle, 0, FILE_BEGIN);
    //    -  
    Buffer:= AllocMem(SHX_FILE_SIGNATURE_SIZE + 1);
    try
      FileRead(FileHandle, Buffer^, SHX_FILE_SIGNATURE_SIZE);
      S:= AnsiString(Buffer);
    finally
      FreeMem(Buffer);
    end;
    FFontKind:= GetFontKind(S);
    if FFontKind = fkUnknown then
      raise Exception.CreateFmt('Error Font Kind [%d]!', [Integer(FFontKind)]);
    if FFontKind = fkBigFont then
      raise Exception.Create('BigFont is not support!');
    //  
    if FFontKind = fkShpFont then
      NumShapesOfs:= SHX_SHAPES_NUM_SHAPES_OFS
    else
      NumShapesOfs:= SHX_UNIFONT_NUM_SHAPES_OFS;
    FileSeek(FileHandle, NumShapesOfs, FILE_BEGIN);
    FileRead(FileHandle, NumShapes, SizeOf(NumShapes));
    //  SHAPE-  : [ ][  (, . )]
    if FFontKind = fkShpFont then
    begin
      for i:= 0 to NumShapes - 1 do
      begin
        FileRead(FileHandle, ShapeInfo, SizeOf(TShxShapeInfo));
        Add(TShxShape.Create(Self, ShapeInfo.ShapeNumber, ShapeInfo.GlyphBytes));
      end;
    end;
    //    
    for i:= 0 to NumShapes - 1 do
    begin
      case FFontKind of
        fkShpFont: LoadShape(FileHandle, i);
        fkUniFont: begin
          FileRead(FileHandle, ShapeInfo, SizeOf(TShxShapeInfo));
          Add(TShxShape.Create(Self, ShapeInfo.ShapeNumber, ShapeInfo.GlyphBytes));
          LoadShape(FileHandle, i);
        end;
      end;
    end;
    if Count < 1 then
      raise Exception.CreateFmt('Error Reading Shapes [%d]!', [Count]);
    //    
    if Shapes[0].Number = 0 then //   
    begin
      FIsFont:= True;
      //      0
      if FFontKind = fkShpFont then
        Shape0GlyphLen:= SHX_SHAPES_SHAPE0_GLYPH_LENGTH
      else
        Shape0GlyphLen:= SHX_UNIFNT_SHAPE0_GLYPH_LENGTH;
      if Shapes[0].GlyphLength < Shape0GlyphLen then
        raise Exception.CreateFmt('Error Shape 0 Corrupt [%d, %d]!',
          [Shapes[0].GlyphLength, Shape0GlyphLen]);
      //    fkShpFont   fkUniFont
      FAbove:= Shapes[0].Glyph[0];
      FBelow:= Shapes[0].Glyph[1];
      FModes:= Shapes[0].Glyph[2];
      //   fkUniFont
      if FFontKind = fkUniFont then
      begin
        FUniEncoding:= Shapes[0].Glyph[3];
        FUniType:= Shapes[0].Glyph[4];
      end;
      if Shapes[0].Name <> '' then
        FFontName:= Shapes[0].Name
      else
        FFontName:= ChangeFileExt(FFontFileName, '');
      //      0,   
      Delete(0);
    end else //   
      SetDefault;
    FFontFileName:= AFileName;
    Result:= True;
  finally
    FileClose(FileHandle);
  end;
end;

initialization
  FontShx:= TShxFont.Create;

finalization
  FontShx.Free;

end.
 